/*
 * Copyright (c) 1997, 1998, 1999
 *      Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *      $Id: sdla_ng.c,v 1.7 2005/01/04 17:43:14 sangoma Exp $
 */


/*
 ****************************************************************************
 * sdla_ng.c   WANPIPE(tm) Multiprotocol WAN Link Driver.  Main module.
 *
 * Author:      Alex Feldman    <al.feldman@sangoma.com>
 *
 * ===========================================================================
 * Sep 10, 2003 Alex Feldman      Initial version.
 ****************************************************************************
*/
#include <net/wanpipe_version.h>
#include <net/wanpipe_includes.h>
#include <net/sdlapci.h>
#include <net/wanpipe.h>        /* WANPIPE common user API definitions */

#if defined(NETGRAPH)
# define NG_WAN_NODE_TYPE        "wp"
/* Netgraph hooks */
# define NG_WAN_HOOK_DEBUG       "debug"
# define NG_WAN_HOOK_RAW         "rawdata"

static  void    wan_ng_watchdog_frame(void * arg);
static  int     wan_ng_init_type(sdla_t* card, void* ignored);

static ng_constructor_t wan_ng_constructor;
static ng_rcvmsg_t      wan_ng_rcvmsg;
static ng_shutdown_t    wan_ng_shutdown;
static ng_newhook_t     wan_ng_newhook;
/*static ng_findhook_t  wan_ng_findhook; */
static ng_connect_t     wan_ng_connect;
static ng_rcvdata_t     wan_ng_rcvdata;
static ng_disconnect_t  wan_ng_disconnect;

static struct ng_type typestruct = {
# if (NG_VERSION < 6)
        NG_VERSION,
# else
        NG_ABI_VERSION,
# endif
        NG_WAN_NODE_TYPE,
        NULL,
        wan_ng_constructor,
        wan_ng_rcvmsg,
# if (NG_VERSION >= 7)
	NULL,
# endif
        wan_ng_shutdown,
        wan_ng_newhook,
        NULL,
        wan_ng_connect,
        wan_ng_rcvdata,
# if (NG_VERSION < 5)
        wan_ng_rcvdata,
#endif
        wan_ng_disconnect,
        NULL
};
MODULE_DEPEND(wanpipe, netgraph, NG_ABI_VERSION,NG_ABI_VERSION,NG_ABI_VERSION);

static int wan_ng_done_init = 0;
#endif


#if defined(NETGRAPH)
int wan_ng_init(sdla_t* card)
{
#if defined(NETGRAPH)
	int err;
	/*
	 * we have found a node, make sure our 'type' is availabe.
	 */
	if (wan_ng_init_type(card, NULL)){
		return -EINVAL;
	}
	if ((err = ng_make_node_common(&typestruct, &card->node)) != 0){
		DEBUG_EVENT("%s: Failed to create NETGRAPH node (err=%d)!\n",
                                  card->devname, err);
		return -EINVAL;
	}
	NG_NODE_SET_PRIVATE(card->node, card);
	card->xmitq.ifq_maxlen = IFQ_MAXLEN;
	card->xmitq_hipri.ifq_maxlen = IFQ_MAXLEN;
# if (NG_VERSION >= 6)
	mtx_init(&card->xmitq.ifq_mtx, "wp_xmitq", NULL, MTX_DEF);
	mtx_init(&card->xmitq_hipri.ifq_mtx, "wp_xmitq_hipri", NULL, MTX_DEF);
#endif
	if ((err = ng_name_node(card->node, card->devname))) {
		DEBUG_EVENT("%s: Failed to assign a netgraph node a name (err = %d)!\n",
                           card->devname, err);
# if (NG_VERSION < 5)
		ng_rmnode(card->node);
# else
		ng_rmnode_self(card->node);
# endif
		NG_NODE_UNREF(card->node);
		card->node = NULL;
		return -EINVAL;
	}
	return 0;
#else
	return -EINVAL;
#endif  /* NETGRAPH */
}

int wan_ng_remove(sdla_t* card)
{
#if defined(NETGRAPH)
	if (card->node){
# if (NG_VERSION < 5)
		ng_rmnode(card->node);
		ng_unname(card->node);
# else
		ng_rmnode_self(card->node);
# endif
		NG_NODE_UNREF(card->node);
		card->node = NULL;
	}
	return 0;
#else
	return -EINVAL;
#endif
}

/*****************************************
 * Device timeout/watchdog routine.
 * called once per second.
 * checks to see that if activity was expected, that it hapenned.
 * At present we only look to see if expected output was completed. */
static void wan_ng_watchdog_frame(void * arg)
{
        return;
}

/***********************************************************************
 * This section contains the methods for the Netgraph interface
 ***********************************************************************/
/*
 * It is not possible or allowable to create a node of this type.
 * If the hardware exists, it will already have created it. */
# if (NG_VERSION < 5)
static int wan_ng_constructor(node_p *nodep)
# else
static int wan_ng_constructor(node_p node)
# endif
{
        return (EINVAL);
}

/*
 * give our ok for a hook to be added...
 * If we are not running this should kick the device into life.
 * The hook's private info points to our stash of info about that
 * channel. */
static int wan_ng_newhook(node_p node, hook_p hook, const char *name)
{
        sdla_t* card = NG_NODE_PRIVATE(node);

        /*
         * check if it's our friend the debug hook
         */
        if (strcmp(name, NG_WAN_HOOK_DEBUG) == 0) {
                NG_HOOK_SET_PRIVATE(hook, NULL);
                card->debug_hook = hook;
                return (0);
        }

        /*
         * Check for raw mode hook.
         */
        if (strcmp(name, NG_WAN_HOOK_RAW) != 0) {
                return (EINVAL);
        }
        card->hook = hook;
        NG_HOOK_SET_PRIVATE(hook, card);
        card->datahooks++;
        if (card->wan_up) card->wan_up(card);

        return (0);
}



/*
 * incoming messages.
 * Just respond to the generic TEXT_STATUS message
 */
# if (NG_VERSION < 5)
static  int
wan_ng_rcvmsg(node_p node,
        struct ng_mesg *msg, const char *retaddr, struct ng_mesg **resp)
# else
static  int
wan_ng_rcvmsg(node_p node, item_p item, hook_p lasthook)
# endif
{
        sdla_t* card = NG_NODE_PRIVATE(node);
        struct ng_mesg* resp_msg = NULL;
        int error = 0;
# if (NG_VERSION >= 5)
        struct ng_mesg* msg;
        NGI_GET_MSG(item, msg);
# endif

        switch (msg->header.typecookie) {
        case NGM_GENERIC_COOKIE:
                switch(msg->header.cmd) {
                case NGM_TEXT_STATUS:
                        {
                        char    *arg;
                        int pos = 0;
                        int resplen;

                        resplen = sizeof(struct ng_mesg) + NG_TEXTRESPONSE;
                        NG_MKRESPONSE(resp_msg, msg, resplen, M_NOWAIT);
                        if (resp_msg == NULL) {
                                error = ENOMEM;
                                break;
                        }
                        arg = (resp_msg)->data;

                        /*
                         * Put in the throughput information.
                         */
                        pos = sprintf(arg, "%ld bytes in, %ld bytes out\n",
                                        card->inbytes, card->outbytes);
                        pos += sprintf(arg + pos,
                                "%ld output errors\n",
                                        card->oerrors);
                        pos += sprintf(arg + pos,
                                "ierrors = %ld", card->ierrors);
                        resp_msg->header.arglen = pos+1;
                        }
                        break;
               default:
                        error = EINVAL;
                        break;
                }
                break;
        default:
                error = EINVAL;
                break;
        }
# if (NG_VERSION < 5)
        *resp = resp_msg;
# else
        NG_RESPOND_MSG(error, node, item, resp_msg);
# endif
        NG_FREE_MSG(msg);
        return (error);
}

/*
 * get data from another node and transmit it to the correct channel
 */
# if (NG_VERSION < 5)
static int wan_ng_rcvdata(hook_p hook, struct mbuf *m, meta_p meta)
# else
static int wan_ng_rcvdata(hook_p hook, item_p item)
# endif
{
        node_p		node = NG_HOOK_NODE(hook);
        sdla_t*		card = NG_NODE_PRIVATE(node);
	netdevice_t*	ifp = NULL;
#ifdef ALTQ
        struct ifaltq*  xmitq_p;
#else
        struct ifqueue* xmitq_p;
#endif
        int             s;
        int             error = 0;
# if (NG_VERSION >= 5)
        struct mbuf*    m;
        meta_p          meta;
        NGI_GET_M(item, m);
        NGI_GET_META(item, meta);
        NG_FREE_ITEM(item);
# endif

        /*
         * data doesn't come in from just anywhere (e.g control hook)
         */
        if (NG_HOOK_PRIVATE(hook) == NULL) {
                error = ENETDOWN;
                goto bad;
        }
        if (card->wan_start == NULL){
		ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
		if (ifp == NULL || ifp->if_start == NULL){
                	error = EINVAL;
                	goto bad;
		}
        }
        /*
         * Now queue the data for when it can be sent
         */
# if (NG_VERSION >= 5) && (NG_VERSION < 7)
        if (meta && meta->priority > 0) {
                xmitq_p = (&card->xmitq_hipri);
        } else 
# endif
	{
		if (card->wan_start){
                	xmitq_p = (&card->xmitq);
		}else{
			xmitq_p = &ifp->if_snd;
		}
        }
        s = splimp();
     	IF_LOCK(xmitq_p);
	if (IF_QFULL(xmitq_p)) {
                IF_DROP(xmitq_p);
     		IF_UNLOCK(xmitq_p);
                splx(s);
                error = ENOBUFS;
                goto bad;
        }
        IF_ENQUEUE(xmitq_p, m);
    	IF_UNLOCK(xmitq_p);
        if (card->wan_start){
		card->wan_start(card);
	} else if (ifp->if_start){
		ifp->if_start(ifp);
	}
        splx(s);
        return (0);

bad:
        /*
         * It was an error case.
         * check if we need to free the mbuf, and then return the error
         */
# if (NG_VERSION < 5)
        NG_FREE_DATA(m, meta);
# else
        NG_FREE_M(m);
        NG_FREE_META(meta);
#endif
        return (error);
}


/*
 * do local shutdown processing..
 * this node will refuse to go away, unless the hardware says to..
 * don't unref the node, or remove our name. just clear our links up.
 */
static int wan_ng_shutdown(node_p node)
{
        sdla_t* card = NG_NODE_PRIVATE(node);

        if (card->wan_down) card->wan_down(card);
# if (NG_VERSION < 5)
        ng_cutlinks(node);
        node->flags &= ~NG_INVALID; /* bounce back to life */
# endif
        return (0);
}

/* already linked */
static int wan_ng_connect(hook_p hook)
{
        /* be really amiable and just say "YUP that's OK by me! " */
        return (0);
}

/*
 * Notify on hook disconnection (destruction)
 *
 * Invalidate the private data associated with this dlci.
 * For this type, removal of the last link resets tries to destroy the node.
 * As the device still exists, the shutdown method will not actually
 * destroy the node, but reset the device and leave it 'fresh' :)
 *
 * The node removal code will remove all references except that owned by the
 * driver.
 */
static int wan_ng_disconnect(hook_p hook)
{
        node_p  node = NG_HOOK_NODE(hook);
        sdla_t* card = NG_NODE_PRIVATE(node);
        int     s;
        /*
         * If it's the data hook, then free resources etc.
         */
        if (NG_HOOK_PRIVATE(hook)) {
                s = splimp();
                card->datahooks--;
                if (card->datahooks == 0){
                        if (card->wan_down) card->wan_down(card);
                }
                splx(s);
        } else {
                card->debug_hook = NULL;
        }

        return (0);
}

/*
 * called during bootup
 * or LKM loading to put this type into the list of known modules
 */
static int wan_ng_init_type(sdla_t* card, void *ignored)
{
        int err = 0;

        WAN_ASSERT(card->wan_down == NULL);
        WAN_ASSERT(card->wan_up == NULL);
        WAN_ASSERT(card->wan_start == NULL);
        if (wan_ng_done_init != 0){
                return 0;
        }

        if (ng_newtype(&typestruct) == EINVAL){
                DEBUG_EVENT("%s: Failed to install a new netgraph type (%d)!\n",
                                        card->devname, err);
                return -EINVAL;
        }
        wan_ng_done_init = 1;
        return 0;
}

#endif
